import ENTITY_FORM_LIST from 'constant/form_template/entity_form_template';
import useDrupalJsonApi from 'hooks/drupal_jsonapi';
import EntityData from 'model/EntityData';
import EntityFormTemplate from 'model/EntityFormTemplate';
import EntityFormLabel from 'molecules/EntityFormLabel';
import EntityFormPanel from 'organisms/EntityFormPanel';
import SideBar from 'organisms/SideBar';
import LoadingDataPage from 'pages/LoadingDataPage';
import { useEffect, useState } from 'react';
import { Button, Col, Container, Form, Row } from 'react-bootstrap';
import { useTranslation } from 'react-i18next';
import { BsCheckLg } from 'react-icons/bs';
import { useHistory, useParams } from 'react-router-dom';
import { capitalize, formDataToSendData } from 'service/utility';

const EntityXxudPageImpl = ({ cloudContext, entityFormTemplate, entityName, entityId }: {
  cloudContext: string,
  entityFormTemplate: EntityFormTemplate,
  entityName: string,
  entityId: string
}) => {
  const [formData, setFormData] = useState<Record<string, any>>({});
  const [loadingFlg, setLoadingFlg] = useState(true);
  const [entityData, setEntityData] = useState<EntityData | undefined>();
  const { getJsonData } = useDrupalJsonApi();
  const { t } = useTranslation();
  const history = useHistory();

  // Initialize form data.
  useEffect(() => {
    setLoadingFlg(true);
    initForm().then(() => {
      setLoadingFlg(false);
    });
    // eslint-disable-next-line react-hooks/exhaustive-deps
  }, []);

  useEffect(() => {
    console.log(formData);
  }, [formData]);

  const initForm = async () => {
    const entityTypeId = `${entityFormTemplate.cloudServiceProvider}_${entityName}`;
    const url = `/jsonapi/${entityTypeId}/${entityTypeId}?filter[drupal_internal__id]=${entityId}`;
    const rawEntityData = await getJsonData<{ data: EntityData[] }>(url);
    if (rawEntityData.data.length >= 1) {
      setEntityData(rawEntityData.data[0]);
    }

    const newFormData: Record<string, any> = {};
    entityFormTemplate.entityRecords.forEach((entityRecord) => {
      if (entityRecord.type === 'panel') {
        entityRecord.keyValueRecords.forEach((keyValueRecord) => {
          if (keyValueRecord.type === 'time') {
            if (rawEntityData.data.length >= 1 && keyValueRecord.hourName in rawEntityData.data[0].attributes) {
              // keyValueRecord.type is 'time' and this entity has value
              newFormData[keyValueRecord.hourName] = rawEntityData.data[0].attributes[keyValueRecord.hourName];
              newFormData[keyValueRecord.minuteName] = rawEntityData.data[0].attributes[keyValueRecord.minuteName];
            } else {
              // keyValueRecord.type is 'time' and this entity has not value
              newFormData[keyValueRecord.hourName] = keyValueRecord.defaultValue[0];
              newFormData[keyValueRecord.minuteName] = keyValueRecord.defaultValue[1];
            }
          } else {
            if (rawEntityData.data.length >= 1
              && keyValueRecord.name in rawEntityData.data[0].attributes
              && rawEntityData.data[0].attributes[keyValueRecord.name] !== null) {
              // keyValueRecord.type is not 'time' and this entity has value
              newFormData[keyValueRecord.name] = rawEntityData.data[0].attributes[keyValueRecord.name];

              // Correction for cases where the value to be obtained is a boolean type.
              if (typeof newFormData[keyValueRecord.name] === 'boolean') {
                newFormData[keyValueRecord.name] = newFormData[keyValueRecord.name] ? '1' : '0';
              }
            } else if (keyValueRecord.type === 'datetime') {
              // The 'datetime' type is read-only so do not include it in the form data.
            }  else {
              // keyValueRecord.type is not 'time' and this entity has not value
              newFormData[keyValueRecord.name] = keyValueRecord.defaultValue;
            }
          }
        });
      }
    });
    setFormData(newFormData);
  }

  const save = async () => {
    const url = `/cloud_dashboard/${entityFormTemplate.cloudServiceProvider}/${cloudContext}/${entityFormTemplate.cloudServiceProvider}_${entityName}/${entityId}/${entityFormTemplate.actionType}`;

    const response = await fetch(url, {
      method: 'POST',
      body: formDataToSendData(formData),
    });
    if (!response.ok) {
      console.log('Create Request : No');
      console.error(capitalize(entityFormTemplate.actionType) + ' failed.');
      console.error(await response.json());
      window.alert(capitalize(entityFormTemplate.actionType) + ' failed.');
    } else {
      console.log('Create Request : Yes');
      console.log(await response.json());

      history.push(`/${entityFormTemplate.cloudServiceProvider}/${cloudContext}/${entityName}`);
    }
  }

  return loadingFlg
    ? <LoadingDataPage />
    : <>
      <SideBar />
      <Container fluid className="px-0">
        <Row className="mx-0">
          <Col>
            <Form>
              {
                entityFormTemplate.entityRecords.map((entityRecord) => {
                  switch (entityRecord.type) {
                    case 'panel':
                      return <EntityFormPanel
                        cloudServiceProvider={entityFormTemplate.cloudServiceProvider}
                        cloudContext={cloudContext}
                        entityName={entityName}
                        actionType={entityFormTemplate.actionType}
                        entityRecord={entityRecord}
                        formData={formData}
                        setFormData={setFormData} />;
                    case 'label':
                      return <EntityFormLabel label={entityRecord.text} entityName={entityName} entityData={entityData} />;
                    default:
                      return <></>;
                  }
                })
              }
              <Button onClick={save}>
                <BsCheckLg />{' ' + t(
                  entityFormTemplate.submitButtonLabel
                    ? entityFormTemplate.submitButtonLabel
                    : 'Save'
                )}
              </Button>
            </Form>
          </Col>
        </Row>
      </Container>
    </>;
}

const EntityXxudPage = () => {
  const params: {
    cloudServiceProvider: string,
    cloudContext: string,
    entityName: string,
    entityId: string,
    action: string
  } = useParams();

  const entityFormTemplate = ENTITY_FORM_LIST.filter((template) => {
    return template.cloudServiceProvider === params.cloudServiceProvider
      && (template.entityName === params.entityName || template.entityName === '')
      && template.actionType === params.action
  });

  if (entityFormTemplate.length >= 1) {
    return <EntityXxudPageImpl
      cloudContext={params.cloudContext}
      entityFormTemplate={entityFormTemplate[0]}
      entityName={params.entityName}
      entityId={params.entityId} />;
  }

  // Default Page
  return <>
    <SideBar />
    <Container fluid className="px-0">
      <Row>
        <Col>
          <span>EntityCxudPage</span><br />
          <span>cloudServiceProvider={params.cloudServiceProvider}</span><br />
          <span>cloudContext={params.cloudContext}</span><br />
          <span>entityName={params.entityName}</span><br />
          <span>entityId={params.entityId}</span><br />
          <span>action={params.action}</span>
        </Col>
      </Row>
    </Container>
  </>;
}

export default EntityXxudPage;
